import numpy as np
import PIL.Image
import tensorflow as tf


def download(url, max_dim=None):
    """Download an image and load it as a NumPy array."""
    name = url.split("/")[-1]
    image_path = tf.keras.utils.get_file(name, origin=url)
    img = PIL.Image.open(image_path)
    if max_dim:
        img.thumbnail((max_dim, max_dim))
    return np.array(img)


def calc_loss(img, model):
    """Calculate the DeepDream loss by maximizing activations."""
    img_batch = tf.expand_dims(img, axis=0)
    layer_activations = model(img_batch)
    if len(layer_activations) == 1:
        layer_activations = [layer_activations]

    return tf.reduce_sum([tf.math.reduce_mean(act) for act in layer_activations])



import tensorflow as tf

from dreamify.utils.deep_dream_utils.utils import calc_loss


class DeepDream(tf.Module):
    def __init__(self, model):
        self.model = model

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.int32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
        )
    )
    def __call__(self, img, steps, step_size):
        print("Tracing DeepDream computation graph...")
        loss = tf.constant(0.0)
        for n in tf.range(steps):
            with tf.GradientTape() as tape:
                tape.watch(img)
                loss = calc_loss(img, self.model)
                gradients = tape.gradient(loss, img)
            gradients /= tf.math.reduce_std(gradients) + 1e-8
            img = img + gradients * step_size
            img = tf.clip_by_value(img, -1, 1)

        return loss, img



from dreamify.utils.deep_dream_utils.deep_dream import DeepDream
from dreamify.utils.deep_dream_utils.tiled_gradients import TiledGradients
from dreamify.utils.deep_dream_utils.utils import download

__all__ = [download, DeepDream, TiledGradients]



import tensorflow as tf
from moviepy.video.fx import AccelDecel
from moviepy.video.VideoClip import DataVideoClip

from dreamify.utils.common import deprocess


class ImageToVideoConverter:
    def __init__(self, dimensions, max_frames_to_sample):
        self.dimensions = dimensions
        self.frames_for_vid: list = []
        self.max_frames_to_sample: int = max_frames_to_sample
        self.curr_frame_idx: int = 0

    def add_to_frames(self, frame):
        frame = tf.image.resize(frame, self.dimensions)
        frame = deprocess(frame.numpy())

        self.frames_for_vid.append(frame)
        self.curr_frame_idx += 1

    def continue_framing(self):
        return self.curr_frame_idx < self.max_frames_to_sample - 1

    def to_video(self, output_path, duration, fps=60):
        self.upsample()

        print(f"Number of images to frame: {len(self.frames_for_vid)}")

        vid = DataVideoClip(self.frames_for_vid, lambda x: x, fps=fps)
        vid = AccelDecel(new_duration=duration).apply(vid)
        vid.write_videofile(output_path)

    def upsample(self):
        NUM_FRAMES_TO_INSERT = 60

        new_frames = []

        # Upsample via frame-frame interpolation
        for i in range(len(self.frames_for_vid) - 1):
            frame1 = tf.cast(self.frames_for_vid[i], tf.float32)
            frame2 = tf.cast(self.frames_for_vid[i + 1], tf.float32)

            new_frames.append(self.frames_for_vid[i].numpy())  # Add original frame

            interpolated = self.interpolate_frames(frame1, frame2, NUM_FRAMES_TO_INSERT)
            new_frames.extend(interpolated.numpy())

        new_frames.extend(
            [self.frames_for_vid[-1].numpy()] * 60 * 3
        )  # Lengthen end frame by 3 frames
        self.frames_for_vid = new_frames

    @tf.function
    def interpolate_frames(self, frame1, frame2, num_frames):
        alphas = tf.linspace(0.0, 1.0, num_frames + 2)[1:-1]  # Avoid frames 0 and 1

        frame1 = tf.cast(frame1, tf.float32)
        frame2 = tf.cast(frame2, tf.float32)

        interpolated_frames = (1 - alphas[:, None, None, None]) * frame1 + alphas[
            :, None, None, None
        ] * frame2
        return tf.cast(interpolated_frames, tf.uint8)



import IPython.display as display
import numpy as np
import tensorflow as tf

from dreamify.utils.common import deprocess, show
from dreamify.utils.deep_dream_utils import DeepDream, TiledGradients, download


def deep_dream_simple(img, dream_model, steps=100, step_size=0.01):
    img = tf.keras.applications.inception_v3.preprocess_input(img)
    img = tf.convert_to_tensor(img)

    step_size = tf.convert_to_tensor(step_size)
    steps_remaining = steps
    step = 0
    while steps_remaining:
        if steps_remaining > 100:
            run_steps = tf.constant(100)
        else:
            run_steps = tf.constant(steps_remaining)
        steps_remaining -= run_steps
        step += run_steps

        loss, img = dream_model(img, run_steps, tf.constant(step_size))

        display.clear_output(wait=True)
        show(deprocess(img))
        print("Step {}, loss {}".format(step, loss))

    result = deprocess(img)

    return result


def deep_dream_octaved(img, dream_model, steps_per_octave=100, step_size=0.01):
    OCTAVE_SCALE = 1.30

    img = tf.constant(np.array(img))
    base_shape = tf.shape(img)[:-1]
    float_base_shape = tf.cast(base_shape, tf.float32)

    for n in range(-2, 3):
        new_shape = tf.cast(float_base_shape * (OCTAVE_SCALE**n), tf.int32)

        img = tf.image.resize(img, new_shape).numpy()
        img = deep_dream_simple(
            img=img,
            dream_model=dream_model,
            steps=steps_per_octave,
            step_size=step_size,
        )

    return img


def deep_dream_rolled(
    img,
    get_tiled_gradients,
    steps_per_octave=100,
    step_size=0.01,
    octaves=range(-2, 3),
    octave_scale=1.3,
):
    base_shape = tf.shape(img)
    img = tf.keras.utils.img_to_array(img)
    img = tf.keras.applications.inception_v3.preprocess_input(img)

    initial_shape = img.shape[:-1]
    img = tf.image.resize(img, initial_shape)
    for octave in octaves:
        # Scale the image based on the octave
        new_size = tf.cast(tf.convert_to_tensor(base_shape[:-1]), tf.float32) * (
            octave_scale**octave
        )
        new_size = tf.cast(new_size, tf.int32)
        img = tf.image.resize(img, new_size)

        for step in range(steps_per_octave):
            gradients = get_tiled_gradients(img, new_size)
            img = img + gradients * step_size
            img = tf.clip_by_value(img, -1, 1)

            if step % 10 == 0:
                display.clear_output(wait=True)
                show(deprocess(img))
                print("Octave {}, Step {}".format(octave, step))

    result = deprocess(img)
    return result


def main():

    url = (
        "https://storage.googleapis.com/download.tensorflow.org/"
        "example_images/YellowLabradorLooking_new.jpg"
    )

    original_img = download(url, max_dim=500)
    show(original_img)

    base_model = tf.keras.applications.InceptionV3(
        include_top=False, weights="imagenet"
    )

    names = ["mixed3", "mixed5"]
    layers = [base_model.get_layer(name).output for name in names]

    dream_model = tf.keras.Model(inputs=base_model.input, outputs=layers)

    deepdream = DeepDream(dream_model)

    # Single Octave
    img = deep_dream_simple(
        img=original_img, dream_model=deepdream, steps=100, step_size=0.01
    )

    img = tf.image.resize(img, original_img.shape[:-1])
    img = tf.image.convert_image_dtype(img / 255.0, dtype=tf.uint8)
    show(img)


def main2():
    url = (
        "https://storage.googleapis.com/download.tensorflow.org/"
        "example_images/YellowLabradorLooking_new.jpg"
    )

    original_img = download(url, max_dim=500)
    show(original_img)

    base_model = tf.keras.applications.InceptionV3(
        include_top=False, weights="imagenet"
    )

    names = ["mixed3", "mixed5"]
    layers = [base_model.get_layer(name).output for name in names]

    dream_model = tf.keras.Model(inputs=base_model.input, outputs=layers)

    deepdream = DeepDream(dream_model)

    # Multi-Octave
    img = deep_dream_octaved(
        img=original_img, dream_model=deepdream, steps_per_octave=50, step_size=0.01
    )
    img = tf.image.resize(img, original_img.shape[:-1])
    img = tf.image.convert_image_dtype(img / 255.0, dtype=tf.uint8)
    display.clear_output(wait=True)
    show(img)


def main3():
    url = (
        "https://storage.googleapis.com/download.tensorflow.org/"
        "example_images/YellowLabradorLooking_new.jpg"
    )

    original_img = download(url, max_dim=500)
    show(original_img)

    base_model = tf.keras.applications.InceptionV3(
        include_top=False, weights="imagenet"
    )

    names = ["mixed3", "mixed5"]
    layers = [base_model.get_layer(name).output for name in names]

    dream_model = tf.keras.Model(inputs=base_model.input, outputs=layers)

    # Rolling/Multi-Octave with Tiling
    get_tiled_gradients = TiledGradients(dream_model)
    img = deep_dream_rolled(
        img=original_img,
        get_tiled_gradients=get_tiled_gradients,
        step_size=0.01,
    )
    img = tf.image.resize(img, original_img.shape[:-1])
    img = tf.image.convert_image_dtype(img / 255.0, dtype=tf.uint8)
    display.clear_output(wait=True)
    show(img)


if __name__ == "__main__":
    main()



import random
from enum import Enum

from tensorflow.keras.applications import (
    VGG19,
    ConvNeXtXLarge,
    DenseNet121,
    EfficientNetV2L,
    InceptionResNetV2,
    InceptionV3,
    MobileNetV2,
    ResNet152V2,
    Xception,
)


class ModelType(Enum):
    VGG19 = "vgg19"
    CONVNEXT_XL = "convnext_xl"
    DENSENET121 = "densenet121"
    EFFICIENTNET_V2L = "efficientnet_v2l"
    INCEPTION_RESNET_V2 = "inception_resnet_v2"
    INCEPTION_V3 = "inception_v3"
    RESNET152V2 = "resnet152v2"
    XCEPTION = "xception"
    MOBILENET_V2 = "mobilenet_v2"


model_layer_settings = {
    ModelType.INCEPTION_V3: {
        "mixed4": 1.0,
        "mixed5": 1.5,
        "mixed6": 2.0,
        "mixed7": 2.5,
    },
    ModelType.VGG19: {
        "block5_conv3": 1.0,
        "block5_conv2": 1.5,
        "block4_conv3": 2.0,
        "block3_conv3": 2.5,
    },
    ModelType.DENSENET121: {
        "conv5_block16_1_conv": 1.0,
        "conv4_block24_1_conv": 1.5,
        "conv3_block16_1_conv": 2.0,
        "conv2_block12_1_conv": 2.5,
    },
    ModelType.EFFICIENTNET_V2L: {
        "block7a_project_bn": 1.0,
        "block6a_expand_activation": 1.5,
        "block5a_expand_activation": 2.0,
        "block4a_expand_activation": 2.5,
    },
    ModelType.INCEPTION_RESNET_V2: {
        "mixed_7a": 1.0,
        "mixed_6a": 1.5,
        "mixed_5a": 2.0,
        "mixed_4a": 2.5,
    },
    ModelType.RESNET152V2: {
        "conv5_block3_out": 1.0,
        "conv4_block6_out": 1.5,
        "conv3_block4_out": 2.0,
        "conv2_block3_out": 2.5,
    },
    ModelType.XCEPTION: {
        "block14_sepconv2_act": 1.0,
        "block13_sepconv2_act": 1.5,
        "block12_sepconv2_act": 2.0,
        "block11_sepconv2_act": 2.5,
    },
    ModelType.CONVNEXT_XL: {
        "stage4_block2_depthwise_conv": 1.0,
        "stage3_block2_depthwise_conv": 1.5,
        "stage2_block2_depthwise_conv": 2.0,
        "stage1_block2_depthwise_conv": 2.5,
    },
    ModelType.MOBILENET_V2: {
        "block_16_depthwise": 1.0,
        "block_13_depthwise": 1.5,
        "block_8_depthwise": 2.0,
        "block_5_depthwise": 2.5,
    },
}


def get_layer_settings(model_name: ModelType, layer_settings=None):
    if layer_settings is None:
        layer_settings = model_layer_settings.get(model_name)
        if not layer_settings:
            raise ValueError(f"Unsupported model: {model_name}")
    return layer_settings


def choose_model(model_name: str = None):
    if model_name in ["random", "any"]:
        model_name = random.choice([model.value for model in ModelType])

    model_name_enum = ModelType[model_name.upper()]

    layer_settings = get_layer_settings(model_name_enum)

    match model_name_enum:
        case ModelType.VGG19:
            return VGG19(weights="imagenet", include_top=False), layer_settings
        case ModelType.CONVNEXT_XL:
            return ConvNeXtXLarge(weights="imagenet", include_top=False), layer_settings
        case ModelType.DENSENET121:
            return DenseNet121(weights="imagenet", include_top=False), layer_settings
        case ModelType.EFFICIENTNET_V2L:
            return (
                EfficientNetV2L(weights="imagenet", include_top=False),
                layer_settings,
            )
        case ModelType.INCEPTION_RESNET_V2:
            return (
                InceptionResNetV2(weights="imagenet", include_top=False),
                layer_settings,
            )
        case ModelType.INCEPTION_V3:
            return InceptionV3(weights="imagenet", include_top=False), layer_settings
        case ModelType.RESNET152V2:
            return ResNet152V2(weights="imagenet", include_top=False), layer_settings
        case ModelType.XCEPTION:
            return Xception(weights="imagenet", include_top=False), layer_settings
        case ModelType.MOBILENET_V2:  # New case for MobileNetV2
            return MobileNetV2(weights="imagenet", include_top=False), layer_settings
        case _:
            raise ValueError(f"Invalid model name: {model_name_enum}")



from dreamify.utils.dream_utils.utils import (
    configure_settings,
    gradient_ascent_loop,
    preprocess_image,
)

# , to_video
__all__ = [
    configure_settings,
    preprocess_image,
    gradient_ascent_loop,
    # to_video,
]



from pathlib import Path

import pytest

from dreamify.dream import generate_dream_image


@pytest.mark.filterwarnings(
    "ignore:The structure of `inputs` doesn't match the expected structure"
)
def test_mock_dream():
    img_path = "examples/example0.jpg"
    out_path = "examples/mock_dream.jpg"

    generate_dream_image(
        img_path, output_path=out_path, num_octave=1, iterations=5, learning_rate=15.0
    )
    Path(out_path).unlink(missing_ok=True)



import numpy as np
import PIL.Image
import tensorflow as tf
from IPython import display


def show(img):
    """Display an image."""
    img = np.array(img)
    img = np.squeeze(img)
    display.display(PIL.Image.fromarray(img))


def deprocess(img):
    """Normalize image for display."""
    img = tf.squeeze(img)
    img = 255 * (img + 1.0) / 2.0
    return tf.cast(img, tf.uint8)



import numpy as np
import tensorflow as tf

# from moviepy.video.fx import AccelDecel
# from moviepy.video.VideoClip import DataVideoClip
from tensorflow import keras
from tqdm import trange

from dreamify.utils.configure import Config

config: Config = None


def configure_settings(**kwargs):
    global config
    config = Config(**kwargs)

    return config


def preprocess_image(image_path):
    img = keras.utils.load_img(image_path)
    img = keras.utils.img_to_array(img)
    img = np.expand_dims(img, axis=0)
    img = keras.applications.inception_v3.preprocess_input(img)
    return img


def compute_loss(input_image):
    features = config.feature_extractor(input_image)
    loss = tf.zeros(shape=())
    for name in features.keys():
        coeff = config.layer_settings[name]
        activation = features[name]
        loss += coeff * tf.reduce_mean(tf.square(activation[:, 2:-2, 2:-2, :]))
    return loss


@tf.function
def gradient_ascent_step(image, learning_rate):
    with tf.GradientTape() as tape:
        tape.watch(image)
        loss = compute_loss(image)
    grads = tape.gradient(loss, image)
    grads = tf.math.l2_normalize(grads)
    image += learning_rate * grads
    image = tf.clip_by_value(image, -1, 1)
    return loss, image


def gradient_ascent_loop(image, iterations, learning_rate, max_loss=None):
    global config

    for i in trange(
        iterations, desc="Gradient Ascent", unit="step", ncols=75, mininterval=0.1
    ):
        loss, image = gradient_ascent_step(image, learning_rate)

        if max_loss is not None and loss > max_loss:
            print(
                f"\nTerminating early: Loss ({loss:.2f}) exceeded max_loss ({max_loss:.2f}).\n"
            )
            break

        framer = config.framer

        if config.enable_framing and framer.continue_framing():
            # frame = tf.image.resize(image, config.original_shape)
            # frame = deprocess(frame.numpy())

            # framer.frames_for_vid.append(frame)
            # framer.curr_frame_idx += 1
            framer.add_to_frames(image)

    return image


# def to_video(output_path, duration, fps=60):
#     global config

#     upsample()

#     print(f"Number of images to frame: {len(config.frames_for_vid)}")

#     vid = DataVideoClip(config.frames_for_vid, lambda x: x, fps=fps)
#     vid = AccelDecel(new_duration=duration).apply(vid)
#     vid.write_videofile(output_path)

#     config = Config()  # Reset the configuration


# @tf.function
# def interpolate_frames(frame1, frame2, num_frames):
#     alphas = tf.linspace(0.0, 1.0, num_frames + 2)[1:-1]  # Avoid 0 and 1

#     frame1 = tf.cast(frame1, tf.float32)
#     frame2 = tf.cast(frame2, tf.float32)

#     interpolated_frames = (1 - alphas[:, None, None, None]) * frame1 + alphas[
#         :, None, None, None
#     ] * frame2
#     return tf.cast(interpolated_frames, tf.uint8)


# def upsample():
#     global config
#     NUM_FRAMES_TO_INSERT = 60

#     new_frames = []

#     # Upsample via frame-frame interpolation
#     for i in range(len(config.frames_for_vid) - 1):
#         frame1 = tf.cast(config.frames_for_vid[i], tf.float32)
#         frame2 = tf.cast(config.frames_for_vid[i + 1], tf.float32)

#         new_frames.append(config.frames_for_vid[i].numpy())  # Add original frame

#         interpolated = interpolate_frames(frame1, frame2, NUM_FRAMES_TO_INSERT)
#         new_frames.extend(interpolated.numpy())

#     new_frames.extend(
#         [config.frames_for_vid[-1].numpy()] * 60 * 3
#     )  # Lengthen end frame by 3 seconds
#     config.frames_for_vid = new_frames


__all__ = [
    configure_settings,
    preprocess_image,
    gradient_ascent_loop,
    # to_video,
]



# from pathlib import Path

# import matplotlib.pyplot as plt
# from PIL import Image

# from dreamify.dream import generate_dream_image
# from dreamify.utils.models import ModelType


# def main():
#     image_path = "examples/example0.jpg"
#     output_dir = Path("examples")

#     output_dir.mkdir(parents=True, exist_ok=True)
#     generated_images = []

#     for model_name in ModelType:
#         print(f"Generating dream image with model: {model_name.value}")

#         output_path = output_dir / f"dream_{model_name.value}.png"

#         generate_dream_image(
#             image_path=image_path,
#             output_path=str(output_path),
#             model_name=model_name.value,
#             step=20.0,
#             num_octave=3,
#             octave_scale=1.4,
#             iterations=30,
#             max_loss=15.0,
#             save_video=False,
#             duration=10,
#         )
#         generated_images.append(output_path)
#         print(
#             f"Dream image generated for {model_name.value} and saved to {output_path}"
#         )

#     compare_images(generated_images)


# def compare_images(image_paths):
#     fig, axes = plt.subplots(1, len(image_paths), figsize=(15, 5))

#     if len(image_paths) == 1:
#         axes = [axes]

#     for ax, image_path in zip(axes, image_paths):
#         img = Image.open(image_path)
#         ax.imshow(img)
#         ax.axis("off")
#         ax.set_title(image_path.name)

#     plt.show()


# if __name__ == "__main__":
#     main()



import tensorflow as tf
from tensorflow import keras

from dreamify.utils.models import choose_model


class FeatureExtractor:
    def __init__(self, model_name):
        self.model, self.layer_settings = choose_model(model_name)

        outputs_dict = {
            layer.name: layer.output
            for layer in [
                self.model.get_layer(name) for name in self.layer_settings.keys()
            ]
        }
        self.feature_extractor = keras.Model(
            inputs=self.model.inputs, outputs=outputs_dict
        )

    @tf.function
    def __call__(self, input):
        return self.feature_extractor(input)



# from dataclasses import dataclass, field


# @dataclass
# class Config:
#     feature_extractor: object = None
#     layer_settings: object = None
#     original_shape: object = None
#     enable_framing: bool = False
#     frames_for_vid: list = field(default_factory=list)
#     max_frames_to_sample: int = 0
#     curr_frame_idx: int = 0


########################################################################

from dataclasses import dataclass

from dreamify.lib.image_to_video_converter import ImageToVideoConverter


@dataclass
class Config:
    feature_extractor: object = None
    layer_settings: object = None
    original_shape: object = None
    enable_framing: bool = False
    max_frames_to_sample: int = 0
    framer: ImageToVideoConverter = None

    def __post_init__(self):
        if self.framer is None:
            self.framer = ImageToVideoConverter(
                dimensions=self.original_shape,
                max_frames_to_sample=self.max_frames_to_sample,
            )
